from pyparsing import Word, Combine, Literal, alphanums, Suppress, Optional, LineEnd

urn_word = Word(alphanums + "_$?=%.&,")
ge_metrics_urn = Combine(
    Suppress(Literal("urn:great_expectations:")) +
    Literal("metrics").setResultsName("urn_type") + Suppress(":") +
    urn_word.setResultsName("run_id") + Suppress(":") +
    urn_word.setResultsName("expectation_suite_name") + Suppress(":") +
    urn_word.setResultsName("metric_name") +
    Optional(Suppress(":") + urn_word.setResultsName("metric_kwargs")) +
    Suppress(LineEnd())
)
ge_validations_urn = Combine(
    Suppress(Literal("urn:great_expectations:")) +
    Literal("validations").setResultsName("urn_type") + Suppress(":") +
    urn_word.setResultsName("expectation_suite_name") + Suppress(":") +
    urn_word.setResultsName("metric_name") +
    Optional(Suppress(":") + urn_word.setResultsName("metric_kwargs")) +
    Suppress(LineEnd())
)
ge_stores_urn = Combine(
    Suppress(Literal("urn:great_expectations:")) +
    Literal("stores").setResultsName("urn_type") + Suppress(":") +
    urn_word.setResultsName("store_name") + Suppress(":") +
    urn_word.setResultsName("metric_name") +
    Optional(Suppress(":") + urn_word.setResultsName("metric_kwargs")) +
    Suppress(LineEnd())
)

ge_urn = ge_metrics_urn | ge_validations_urn | ge_stores_urn



import datetime

import pytest

from great_expectations.core import RunIdentifier
from great_expectations.core.metric import ValidationMetricIdentifier
from great_expectations.data_context.util import instantiate_class_from_config


@pytest.fixture(params=[
    {
        "class_name": "MetricStore",
        "store_backend": {
            "class_name": "DatabaseStoreBackend",
            "credentials": {
                "drivername": "postgresql",
                "username": "postgres",
                "password": "",
                "host": "localhost",
                "port": "5432",
                "database": "test_ci"
            }
        }
    },
    {
        "class_name": "MetricStore",
        "module_name": "great_expectations.data_context.store"
    }
])
def param_store(request, test_backends):
    # If we have a backend configuration but we do not have postgres configured, skip
    backend_config = request.param.get("store_backend", None)
    if backend_config:
        if backend_config.get("credentials", {}).get("drivername", None) == "postgresql":
            if "postgresql" not in test_backends:
                pytest.skip("skipping fixture because postgresql not selected")

    return instantiate_class_from_config(
        config=request.param,
        config_defaults={
            "module_name": "great_expectations.data_context.store",
        },
        runtime_environment={}
    )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import json
import os

import mock
from great_expectations.cli import toolkit

@mock.patch("subprocess.call", return_value=True, side_effect=None)
def test_launch_jupyter_notebook_env_none(mock_subprocess):
    try:
        if os.environ.get('GE_JUPYTER_CMD'):
            temp = os.environ.great_expectations('GE_JUPYTER_CMD')
            del os.environ['GE_JUPYTER_CMD']
        toolkit.launch_jupyter_notebook('test_path')
        mock_subprocess.assert_called_once_with(["jupyter", "notebook", "test_path"])
    except Exception:
        if temp:
            os.environ['GE_JUPYTER_CMD'] = temp
        raise

@mock.patch("subprocess.call", return_value=True, side_effect=None)
def test_launch_jupyter_notebook_env_set_in_env(mock_subprocess):
    with mock.patch.dict(os.environ, {'GE_JUPYTER_CMD': 'jupyter notebook --test-args'}, clear=True):
        toolkit.launch_jupyter_notebook('test_path')
        mock_subprocess.assert_called_once_with("jupyter notebook --test-args test_path", shell=True)









import datetime
import os

from dateutil.parser import parse, ParserError

from great_expectations import DataContext
from great_expectations.cli.util import cli_message
from great_expectations.data_context.store import ValidationsStore, HtmlSiteStore, TupleFilesystemStoreBackend, \
    TupleS3StoreBackend, TupleGCSStoreBackend, MetricStore, DatabaseStoreBackend, InMemoryStoreBackend
from great_expectations.data_context.types.resource_identifiers import ValidationResultIdentifier


class UpgradeHelperV11:
    def __init__(self, data_context=None, context_root_dir=None):
        assert data_context or context_root_dir, "Please provide a data_context object or a context_root_dir."

        self.data_context = data_context or DataContext(context_root_dir=context_root_dir)

        self.upgrade_log = {
            "skipped_validations_stores": {
                "database_store_backends": [],
                "unsupported": []
            },
            "skipped_docs_validations_stores": {
                "unsupported": []
            },
            "skipped_metrics_stores": {
                "database_store_backends": [],
                "unsupported": []
            },
            "upgraded_validations_stores": {},
            "upgraded_docs_site_validations_stores": {},
        }

        self.upgrade_checklist = {
            "validations_store_backends": {},
            "docs_validations_store_backends": {}
        }

        self.validation_run_times = {}

        self.run_time_setters_by_backend_type = {
            TupleFilesystemStoreBackend: self._get_tuple_filesystem_store_backend_run_time,
            TupleS3StoreBackend: self._get_tuple_s3_store_backend_run_time,
            TupleGCSStoreBackend: self._get_tuple_gcs_store_backend_run_time
        }

        self.run_id_updaters_by_backend_type = {
            TupleFilesystemStoreBackend: self._update_tuple_filesystem_store_backend_run_id,
            TupleS3StoreBackend: self._update_tuple_s3_store_backend_run_id,
            TupleGCSStoreBackend: self._update_tuple_gcs_store_backend_run_id
        }

    def _generate_upgrade_checklist(self):
        for (store_name, store) in self.data_context.stores.items():
            if not isinstance(store, (ValidationsStore, MetricStore)):
                continue
            elif isinstance(store, ValidationsStore):
                self._process_validations_store_for_checklist(store_name, store)
            elif isinstance(store, MetricStore):
                self._process_metrics_store_for_checklist(store_name, store)

        sites = self.data_context._project_config_with_variables_substituted.data_docs_sites

        if sites:
            for site_name, site_config in sites.items():
                self._process_docs_site_for_checklist(site_name, site_config)

    def _process_docs_site_for_checklist(self, site_name, site_config):
        site_html_store = HtmlSiteStore(
            store_backend=site_config.get("store_backend"),
            runtime_environment={
                "data_context": self.data_context,
                "root_directory": self.data_context.root_directory,
                "site_name": site_name
            }
        )
        site_validations_store_backend = site_html_store.store_backends[ValidationResultIdentifier]

        if isinstance(site_validations_store_backend, tuple(list(self.run_time_setters_by_backend_type.keys()))):
            self.upgrade_checklist["docs_validations_store_backends"][site_name] = site_validations_store_backend
        else:
            self.upgrade_log["skipped_docs_validations_stores"]["unsupported"].append({
                "site_name": site_name,
                "validations_store_backend_class": type(site_validations_store_backend).__name__
            })

    def _process_validations_store_for_checklist(self, store_name, store):
        store_backend = store.store_backend
        if isinstance(store_backend, DatabaseStoreBackend):
            self.upgrade_log["skipped_validations_stores"]["database_store_backends"].append({
                "store_name": store_name,
                "store_backend_class": type(store_backend).__name__
            })
        elif isinstance(store_backend, tuple(list(self.run_time_setters_by_backend_type.keys()))):
            self.upgrade_checklist["validations_store_backends"][store_name] = store_backend
        else:
            self.upgrade_log["skipped_validations_stores"]["unsupported"].append({
                "store_name": store_name,
                "store_backend_class": type(store_backend).__name__
            })

    def _process_metrics_store_for_checklist(self, store_name, store):
        store_backend = store.store_backend
        if isinstance(store_backend, DatabaseStoreBackend):
            self.upgrade_log["skipped_metrics_stores"]["database_store_backends"].append({
                "store_name": store_name,
                "store_backend_class": type(store_backend).__name__
            })
        elif isinstance(store_backend, InMemoryStoreBackend):
            pass
        else:
            self.upgrade_log["skipped_metrics_stores"]["unsupported"].append({
                "store_name": store_name,
                "store_backend_class": type(store_backend).__name__
            })

    # TODO: add logic that changes old run_id to typed run_id in validation json files
    def upgrade_store_backend(self, store_backend):
        validation_source_keys = store_backend.list_keys()

        for source_key in validation_source_keys:
            run_name = source_key[-2]
            if run_name not in self.validation_run_times:
                self.run_time_setters_by_backend_type.get(type(store_backend))(source_key, store_backend)
            dest_key_list = list(source_key)
            dest_key_list.insert(-1, self.validation_run_times[run_name])
            dest_key = tuple(dest_key_list)
            store_backend.move(source_key, dest_key)

    def _update_tuple_filesystem_store_backend_run_id(self):
        pass

    def _update_tuple_s3_store_backend_run_id(self):
        pass

    def _update_tuple_gcs_store_backend_run_id(self):
        pass

    def _get_tuple_filesystem_store_backend_run_time(self, source_key, store_backend):
        run_name = source_key[-2]
        try:
            self.validation_run_times[run_name] = parse(run_name).isoformat()
        except ParserError:
            source_path = os.path.join(
                store_backend.full_base_directory,
                store_backend._convert_key_to_filepath(source_key)
            )
            path_mod_timestamp = os.path.getmtime(source_path)
            path_mod_iso_str = datetime.datetime.fromtimestamp(
                path_mod_timestamp,
                tz=datetime.timezone.utc
            ).isoformat()
            self.validation_run_times[run_name] = path_mod_iso_str

    def _get_tuple_s3_store_backend_run_time(self, source_key, store_backend):
        import boto3
        s3 = boto3.resource('s3')
        run_name = source_key[-2]

        try:
            self.validation_run_times[run_name] = parse(run_name).isoformat()
        except ParserError:
            source_path = store_backend._convert_key_to_filepath(source_key)
            if not source_path.startswith(store_backend.prefix):
                source_path = os.path.join(store_backend.prefix, source_path)
            source_object = s3.Object(store_backend.bucket, source_path)
            source_object_last_mod = source_object.last_modified.isoformat()

            self.validation_run_times[run_name] = source_object_last_mod

    def _get_tuple_gcs_store_backend_run_time(self, source_key, store_backend):
        from google.cloud import storage
        gcs = storage.Client(project=store_backend.project)
        bucket = gcs.get_bucket(store_backend.bucket)
        run_name = source_key[-2]

        try:
            self.validation_run_times[run_name] = parse(run_name).isoformat()
        except ParserError:
            source_path = store_backend._convert_key_to_filepath(source_key)
            if not source_path.startswith(store_backend.prefix):
                source_path = os.path.join(store_backend.prefix, source_path)
            source_blob_created_time = bucket.get_blob(source_path).time_created.isoformat()

            self.validation_run_times[run_name] = source_blob_created_time

    def _generate_upgrade_prompt(self):
        stores_with_database_backends = (self.upgrade_log["skipped_validations_stores"]["database_store_backends"] +
                                         self.upgrade_log["skipped_metrics_stores"]["database_store_backends"])
        unsupported_stores = (self.upgrade_log["skipped_validations_stores"]["unsupported"] +
                              self.upgrade_log["skipped_metrics_stores"]["unsupported"])
        unsupported_doc_sites = self.upgrade_log["skipped_docs_validations_stores"]["unsupported"]

    def upgrade_project(self):
        for (store_name, store_backend) in self.upgrade_checklist["validations_store_backends"].items():
            self.upgrade_store_backend(store_backend)
        for (site_name, store_backend) in self.upgrade_checklist["docs_validations_store_backends"].items():
            self.upgrade_store_backend(store_backend)



import datetime
import json
import logging
import urllib
from collections import OrderedDict

import traceback

from dateutil.parser import parse

from .renderer import Renderer
from great_expectations.render.types import (
    RenderedSectionContent,
    RenderedDocumentContent,
    RenderedHeaderContent,
    RenderedStringTemplateContent,
    RenderedTableContent,
    RenderedBulletListContent,
    RenderedBootstrapTableContent, RenderedTabsContent)
from great_expectations.exceptions import GreatExpectationsError

from .call_to_action_renderer import CallToActionRenderer

logger = logging.getLogger(__name__)


# FIXME : This class needs to be rebuilt to accept SiteSectionIdentifiers as input.
# FIXME : This class needs tests.
class SiteIndexPageRenderer(Renderer):

    @classmethod
    def _generate_expectation_suites_link_table(cls, index_links_dict):
        table_options = {
            "search": "true",
            "trimOnSearch": "false",
            "visibleSearch": "true",
            "rowStyle": "rowStyleLinks",
            "rowAttributes": "rowAttributesLinks",
            "sortName": "expectation_suite_name",
            "sortOrder": "asc",
            "pagination": "true",
            "iconSize": "sm",
            "toolbarAlign": "right"
        }
        table_columns = [
            {
                "field": "expectation_suite_name",
                "title": "Expectation Suites",
                "sortable": "true"
            },
        ]
        expectation_suite_link_dicts = index_links_dict.get("expectations_links", [])
        table_data = []

        for dict_ in expectation_suite_link_dicts:
            table_data.append({
                "expectation_suite_name": dict_.get("expectation_suite_name"),
                "_table_row_link_path": dict_.get("filepath")
            })

        return RenderedBootstrapTableContent(**{
            "table_columns": table_columns,
            "table_data": table_data,
            "table_options": table_options,
            "styling": {
                "classes": ["col-12", "ge-index-page-table-container"],
                "body": {
                    "classes": ["table-sm", "ge-index-page-expectation_suites-table"]
                }
            }
        })

    @classmethod
    def _generate_profiling_results_link_table(cls, index_links_dict):
        table_options = {
            "search": "true",
            "trimOnSearch": "false",
            "visibleSearch": "true",
            "rowStyle": "rowStyleLinks",
            "rowAttributes": "rowAttributesLinks",
            "sortName": "run_time",
            "sortOrder": "desc",
            "pagination": "true",
            "filterControl": "true",
            "iconSize": "sm",
            "toolbarAlign": "right"
        }
        table_columns = [
            {
                "field": "run_time",
                "title": "Run Time",
                "sortName": "_run_time_sort",
                "sortable": "true",
                "filterControl": "datepicker",
            },
            {
                "field": "asset_name",
                "title": "Asset Name",
                "sortable": "true",
                "filterControl": "select"
            },
            {
                "field": "batch_identifier",
                "title": "Batch ID",
                "sortName": "_batch_identifier_sort",
                "sortable": "true",
                "filterControl": "input",
            },
            {
                "field": "profiler_name",
                "title": "Profiler",
                "sortable": "true",
                "filterControl": "select",
            }
        ]
        profiling_link_dicts = index_links_dict.get("profiling_links", [])
        table_data = []

        for dict_ in profiling_link_dicts:
            table_data.append({
                "run_time": cls._get_formatted_datetime(dict_.get("run_time")),
                "_run_time_sort": cls._get_timestamp(dict_.get("run_time")),
                "asset_name": dict_.get("asset_name"),
                "batch_identifier": cls._render_batch_id_cell(
                    dict_.get("batch_identifier"), dict_.get("batch_kwargs")),
                "_batch_identifier_sort": dict_.get("batch_identifier"),
                "profiler_name": dict_.get("expectation_suite_name").split(".")[-1],
                "_table_row_link_path": dict_.get("filepath")
            })

        return RenderedBootstrapTableContent(**{
            "table_columns": table_columns,
            "table_data": table_data,
            "table_options": table_options,
            "styling": {
                "classes": ["col-12", "ge-index-page-table-container"],
                "body": {
                    "classes": ["table-sm", "ge-index-page-profiling-results-table"]
                }
            }
        })

    @classmethod
    def _generate_validation_results_link_table(cls, index_links_dict):
        table_options = {
            "search": "true",
            "trimOnSearch": "false",
            "visibleSearch": "true",
            "rowStyle": "rowStyleLinks",
            "rowAttributes": "rowAttributesLinks",
            "sortName": "run_time",
            "sortOrder": "desc",
            "pagination": "true",
            "filterControl": "true",
            "iconSize": "sm",
            "toolbarAlign": "right"
        }

        table_columns = [
            {
                "field": "validation_success",
                "title": "Status",
                "sortable": "true",
                "align": "center",
                "filterControl": "select",
                "filterDataCollector": "validationSuccessFilterDataCollector",
            },
            {
                "field": "run_time",
                "title": "Run Time",
                "sortName": "_run_time_sort",
                "sortable": "true",
                "filterControl": "datepicker",
            },
            {
                "field": "run_name",
                "title": "Run Name",
                "sortable": "true",
                "filterControl": "input",
            },
            {
                "field": "asset_name",
                "title": "Asset Name",
                "sortable": "true",
                "filterControl": "select"
            },
            {
                "field": "batch_identifier",
                "title": "Batch ID",
                "sortName": "_batch_identifier_sort",
                "sortable": "true",
                "filterControl": "input",
            },
            {
                "field": "expectation_suite_name",
                "title": "Expectation Suite",
                "sortName": "_expectation_suite_name_sort",
                "sortable": "true",
                "filterControl": "select",
                "filterDataCollector": "expectationSuiteNameFilterDataCollector"
            }
        ]
        validation_link_dicts = index_links_dict.get("validations_links", [])
        table_data = []

        for dict_ in validation_link_dicts:
            table_data.append({
                "validation_success": cls._render_validation_success_cell(dict_.get("validation_success")),
                "run_time": cls._get_formatted_datetime(dict_.get("run_time")),
                "_run_time_sort": cls._get_timestamp(dict_.get("run_time")),
                "run_name": dict_.get("run_name"),
                "batch_identifier": cls._render_batch_id_cell(
                    dict_.get("batch_identifier"), dict_.get("batch_kwargs")),
                "_batch_identifier_sort": dict_.get("batch_identifier"),
                "expectation_suite_name": cls._render_expectation_suite_cell(
                    dict_.get("expectation_suite_name"),
                    dict_.get("expectation_suite_filepath")
                ),
                "_expectation_suite_name_sort": dict_.get("expectation_suite_name"),
                "_table_row_link_path": dict_.get("filepath"),
                "_validation_success_text": "Success" if dict_.get("validation_success") else "Failed",
                "asset_name": dict_.get("asset_name")
            })

        return RenderedBootstrapTableContent(**{
            "table_columns": table_columns,
            "table_data": table_data,
            "table_options": table_options,
            "styling": {
                "classes": ["col-12", "ge-index-page-table-container"],
                "body": {
                    "classes": ["table-sm", "ge-index-page-validation-results-table"]
                }
            }
        })

    @classmethod
    def _render_expectation_suite_cell(cls, expectation_suite_name, expectation_suite_path):
        return RenderedStringTemplateContent(**{
            "content_block_type": "string_template",
            "string_template": {
                "template": "$link_text",
                "params": {
                    "link_text": expectation_suite_name
                },
                "tag": "a",
                "styling": {
                    "styles": {
                      "word-break": "break-all"
                    },
                    "attributes": {
                        "href": expectation_suite_path
                    },
                    "classes": ["ge-index-page-table-expectation-suite-link"]
                }
            },
        })

    @classmethod
    def _render_batch_id_cell(cls, batch_id, batch_kwargs):
        return RenderedStringTemplateContent(**{
            "content_block_type": "string_template",
            "string_template": {
                "template": str(batch_id),
                "tooltip": {
                    "content": "Batch Kwargs:\n\n" + json.dumps(batch_kwargs, indent=2),
                    "placement": "top"
                },
                "styling": {
                    "classes": ["m-0", "p-0"]
                }
            }
        })

    @classmethod
    def _get_formatted_datetime(cls, _datetime):
        if isinstance(_datetime, datetime.datetime):
            return _datetime.strftime("%m/%d/%Y %H:%M:%S")
        elif isinstance(_datetime, str):
            return parse(_datetime).strftime("%m/%d/%Y %H:%M:%S")
        else:
            return None

    @classmethod
    def _get_timestamp(cls, _datetime):
        if isinstance(_datetime, datetime.datetime):
            return _datetime.timestamp()
        elif isinstance(_datetime, str):
            return parse(_datetime).timestamp()
        else:
            return ""

    @classmethod
    def _render_validation_success_cell(cls, validation_success):
        return RenderedStringTemplateContent(
            **{
                "content_block_type": "string_template",
                "string_template": {
                    "template": "$validation_success",
                    "params": {
                        "validation_success": ""
                    },
                    "styling": {
                        "params": {
                            "validation_success": {
                                "tag": "i",
                                "classes": ["fas", "fa-check-circle", "text-success", "ge-success-icon"] if
                                    validation_success else ["fas", "fa-times", "text-danger", "ge-failed-icon"]
                            }
                        },
                        "classes": ["ge-index-page-table-validation-links-item"]
                    }
                }
            })

    @classmethod
    def render(cls, index_links_dict):
        sections = []
        cta_object = index_links_dict.pop("cta_object", None)

        try:
            content_blocks = []
            # site name header
            site_name_header_block = RenderedHeaderContent(**{
                "content_block_type": "header",
                "header": RenderedStringTemplateContent(**{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "$title_prefix | $site_name",
                        "params": {
                            "site_name": index_links_dict.get("site_name"),
                            "title_prefix": "Data Docs"
                        },
                        "styling": {
                            "params": {
                                "title_prefix": {
                                    "tag": "strong"
                                }
                            }
                        },
                    }
                }),
                "styling": {
                    "classes": ["col-12", "ge-index-page-site-name-title"],
                    "header": {
                        "classes": ["alert", "alert-secondary"]
                    }
                }
            })
            content_blocks.append(site_name_header_block)

            tabs = []

            if index_links_dict.get("validations_links"):
                tabs.append({
                    "tab_name": "Validation Results",
                    "tab_content": cls._generate_validation_results_link_table(index_links_dict)
                })
            if index_links_dict.get("profiling_links"):
                tabs.append({
                    "tab_name": "Profiling Results",
                    "tab_content": cls._generate_profiling_results_link_table(index_links_dict)
                })
            if index_links_dict.get("expectations_links"):
                tabs.append({
                    "tab_name": "Expectation Suites",
                    "tab_content": cls._generate_expectation_suites_link_table(index_links_dict)
                })

            tabs_content_block = RenderedTabsContent(**{
                "tabs": tabs,
                "styling": {
                    "classes": ["col-12", "ge-index-page-tabs-container"],
                }
            })

            content_blocks.append(tabs_content_block)

            section = RenderedSectionContent(**{
                "section_name": index_links_dict.get("site_name"),
                "content_blocks": content_blocks
            })
            sections.append(section)

            index_page_document = RenderedDocumentContent(**{
                "renderer_type": "SiteIndexPageRenderer",
                "utm_medium": "index-page",
                "sections": sections
            })

            if cta_object:
                index_page_document.cta_footer = CallToActionRenderer.render(cta_object)

            return index_page_document

        except Exception as e:
            exception_message = f'''\
An unexpected Exception occurred during data docs rendering.  Because of this error, certain parts of data docs will \
not be rendered properly and/or may not appear altogether.  Please use the trace, included in this message, to \
diagnose and repair the underlying issue.  Detailed information follows:
            '''
            exception_traceback = traceback.format_exc()
            exception_message += f'{type(e).__name__}: "{str(e)}".  Traceback: "{exception_traceback}".'
            logger.error(exception_message, e, exc_info=True)
